Package: diffHic
Version: 1.44.0
Date: 2024-12-16
Title: Differential Analysis of Hi-C Data
Author: Aaron Lun, Gordon Smyth
Maintainer: Aaron Lun <infinite.monkeys.with.keyboards@gmail.com>, Gordon Smyth <smyth@wehi.edu.au>, Hannah Coughlin <coughlin.h@wehi.edu.au>
Depends: R (>= 3.5), GenomicRanges, InteractionSet,
        SummarizedExperiment
Imports: Rsamtools, Rhtslib, Biostrings, BSgenome, rhdf5, edgeR, limma,
        csaw, locfit, methods, IRanges, S4Vectors, GenomeInfoDb,
        BiocGenerics, grDevices, graphics, stats, utils, Rcpp,
        rtracklayer
Suggests: BSgenome.Ecoli.NCBI.20080805, Matrix, testthat
biocViews: MultipleComparison, Preprocessing, Sequencing, Coverage,
        Alignment, Normalization, Clustering, HiC
Description: Detects differential interactions across biological
        conditions in a Hi-C experiment. Methods are provided for read
        alignment and data pre-processing into interaction counts.
        Statistical analysis is based on edgeR and supports
        normalization and filtering. Several visualization options are
        also available.
License: GPL-3
NeedsCompilation: yes
LinkingTo: Rhtslib (>= 1.13.1), Rcpp
SystemRequirements: C++, GNU make
git_url: https://git.bioconductor.org/packages/diffHic
git_branch: RELEASE_3_23
git_last_commit: b22054b
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
Packaged: 2026-04-28 21:46:07 UTC; biocbuild
Built: R 4.6.1; x86_64-pc-linux-gnu; 2026-08-13 20:16:55 UTC; unix
