#  Calling the dynamic library.
useDynLib(diffHic, .registration=TRUE, .fixes="cxx_")

#  All functions exported, sorted by approximate function:
export(diffHicUsersGuide,
	cutGenome, segmentGenome, emptyGenome,
    preparePairs, mergePairs, savePairs, prepPseudoPairs,
	getPairData, prunePairs, 
	loadChromos, loadData, extractPatch,
	pairParam,
	readMTX2IntSet,
    squareCounts, connectCounts, marginCounts, totalCounts, 
    mergeCMs,
    correctedContact, normalizeCNV, matchMargins,
    getArea,
	filterDirect, filterTrended, filterDiag,
	filterPeaks, enrichedPairs, neighborCounts,
    boxPairs, clusterPairs, consolidatePairs, diClusters,
    annotatePairs,
	compartmentalize, domainDirections,
    plotPlaid, rotPlaid, plotDI, rotDI,
	asDGEList, reform # From csaw, but also available to the user.
)

# Exporting classes.
exportClasses(
    pairParam
)

# Requires several imports.
import(Rsamtools,Rhtslib,GenomicRanges,Biostrings,BSgenome,rhdf5,
	IRanges,S4Vectors,GenomeInfoDb,
	edgeR,limma,locfit,methods,csaw,
    InteractionSet,SummarizedExperiment
)
	
importFrom("BiocGenerics", counts)
importFrom("grDevices", col2rgb, rgb)
importFrom("graphics", box, par, plot, polygon, rect, text)
importFrom("stats", approx, approxfun, fitted, kmeans)
importFrom("utils", read.table, write.table)
importFrom("Rcpp", sourceCpp)

importFrom("rtracklayer", import.bed)
