# Differential analyses of Hi-C data

|Environment|Status|
|---|---|
|[BioC-release](https://bioconductor.org/packages/release/bioc/html/diffHic.html)|[![Release OK](https://bioconductor.org/shields/build/release/bioc/diffHic.svg)](http://bioconductor.org/checkResults/release/bioc-LATEST/diffHic/)|
|[BioC-devel](https://bioconductor.org/packages/devel/bioc/html/diffHic.html)|[![Devel OK](https://bioconductor.org/shields/build/devel/bioc/diffHic.svg)](http://bioconductor.org/checkResults/devel/bioc-LATEST/diffHic/)|

**diffHic** implements methods for detecting differential interactions across biological conditions in a Hi-C experiment.
Methods are provided for read alignment and data pre-processing into interaction counts, with the subsequent statistics analysis performed by **edgeR**.
Check out the [Bioconductor landing page](https://bioconductor.org/packages/diffHic) for more details.
